
@import "CPArray.j"
@import "CPNull.j"
@import "CPString.j"
@import "CPEnumerator.j"

@import "CPPredicate.j"
@import "CPExpression.j"
@import "CPExpression_operator.j"

CPDirectPredicateModifier = 0;
CPAllPredicateModifier = 1;
CPAnyPredicateModifier = 2;

CPLessThanPredicateOperatorType = 0;
CPLessThanOrEqualToPredicateOperatorType = 1;
CPGreaterThanPredicateOperatorType = 2;
CPGreaterThanOrEqualToPredicateOperatorType = 3;
CPEqualToPredicateOperatorType = 4;
CPNotEqualToPredicateOperatorType = 5;
CPMatchesPredicateOperatorType = 6;
CPLikePredicateOperatorType = 7;
CPBeginsWithPredicateOperatorType = 8;
CPEndsWithPredicateOperatorType = 9;
CPInPredicateOperatorType = 10;
CPCustomSelectorPredicateOperatorType = 11;
CPContainsPredicateOperatorType = 12;
CPBetweenPredicateOperatorType = 13;

CPCaseInsensitivePredicateOption = 1;
CPDiacriticInsensitivePredicateOption = 2;

var CPComparisonPredicateModifier;
var CPPredicateOperatorType;

@implementation CPString (Compare)

- (CPComparisonResult)compare:(CPString)aString options:(int)aMask range:(CPRange)range
{   
    if (range.location + range.length > [self length])
        [CPException raise:CPRangeException reason:"Range exceeds the length of the receiver"];
                
    var lhs = [self substringWithRange:range],
        rhs = aString;
        
    return [lhs compare:rhs options:aMask];
}

@end

@implementation CPComparisonPredicate : CPPredicate
{
   CPExpression _left;
   CPExpression _right;
   
   CPComparisonPredicateModifier _modifier;
   CPPredicateOperatorType _type;
   unsigned int _options;
   SEL _customSelector;
}

- (id)initWithLeftExpression:(CPExpression)left rightExpression:(CPExpression)right modifier:(CPComparisonPredicateModifier)modifier type:(CPPredicateOperatorType)type options:(unsigned)options
{
    _left = left;
    _right = right;
    _modifier = modifier;
    _type = type;
    _options = options;
    
    if (type != CPMatchesPredicateOperatorType &&
        type != CPLikePredicateOperatorType &&
        type != CPBeginsWithPredicateOperatorType &&
        type != CPEndsWithPredicateOperatorType &&
        type != CPInPredicateOperatorType)
    _options = 0;

    _customSelector = 0;
    
    return self;
}

- (id)initWithLeftExpression:(CPExpression)left rightExpression:(CPExpression)right customSelector:(SEL)selector
{
    _left = left;
    _right = right;
    _modifier = CPDirectPredicateModifier;
    _type = CPCustomSelectorPredicateOperatorType;
    _options = 0;
    _customSelector = selector;
    
    return self;
}

- (id)initWithCoder:(CPCoder)coder
{
    // UNIMPLEMENTED
    return self;
}

- (void)encodeWithCoder:(CPCoder)coder
{
    // UNIMPLEMENTED
}

+ (CPPredicate)predicateWithLeftExpression:(CPExpression)left rightExpression:(CPExpression)right modifier:(CPComparisonPredicateModifier)modifier type:(int)type options:(unsigned)options
{
    return [[self alloc] initWithLeftExpression:left rightExpression:right modifier:modifier type:type options:options];
}

+ (CPPredicate)predicateWithLeftExpression:(CPExpression)left rightExpression:(CPExpression)right customSelector:(SEL)selector
{
    return [[self alloc] initWithLeftExpression:left rightExpression:right customSelector:selector];
}

- (CPString)predicateFormat
{
    var modifier;

    switch (_modifier)
    {
        case CPDirectPredicateModifier:
            modifier = "";
            break;

        case CPAllPredicateModifier:
            modifier = "ALL ";
            break;

        case CPAnyPredicateModifier:
            modifier = "ANY ";
            break;

        default:
            modifier = "";
            break;
    }

    var options;
    
    switch (_options)
    {
          case CPCaseInsensitivePredicateOption:
            options = "[c]";
            break;
                
          case CPDiacriticInsensitivePredicateOption:
            options = "[d]";
            break;
            
          case CPCaseInsensitivePredicateOption | CPDiacriticInsensitivePredicateOption:
            options = "[cd]";
            break;
            
          default:
            options = "";
            break;
    }

    var operator;

    switch (_type)
    {
        case CPLessThanPredicateOperatorType:
            operator = "<";
            break;

        case CPLessThanOrEqualToPredicateOperatorType:
            operator = "<=";
            break;

        case CPGreaterThanPredicateOperatorType:
            operator = ">";
            break;

        case CPGreaterThanOrEqualToPredicateOperatorType:
            operator = ">=";
            break;

        case CPEqualToPredicateOperatorType:
            operator = "==";
            break;

        case CPNotEqualToPredicateOperatorType:
            operator = "!=";
            break;

        case CPMatchesPredicateOperatorType:
            operator = "MATCHES";
            break;

        case CPLikePredicateOperatorType:
            operator = "LIKE";
            break;

        case CPBeginsWithPredicateOperatorType:
            operator = "BEGINSWITH";
            break;

        case CPEndsWithPredicateOperatorType:
            operator = "ENDSWITH";
            break;

        case CPInPredicateOperatorType:
            operator = "IN";
            break;

            // FIX, not right
        case CPCustomSelectorPredicateOperatorType:
            operator = CPStringFromSelector(_customSelector);
            break;
    }

    return [CPString stringWithFormat:@"%s%s %s%s %s",modifier,[_left description],operator,options,[_right description]];

}

- (CPPredicate)predicateWithSubstitutionVariables:(CPDictionary)variables
{
    var left = [_left _expressionWithSubstitutionVariables:variables];
    var right = [_right _expressionWithSubstitutionVariables:variables];

    if (_type != CPCustomSelectorPredicateOperatorType)
        return [CPComparisonPredicate predicateWithLeftExpression:left rightExpression:right modifier:_modifier type:_type options:_options];
    else
        return [CPComparisonPredicate predicateWithLeftExpression:left rightExpression:right customSelector:_customSelector];
}

- (CPExpression)leftExpression
{
    return _left;
}

- (CPExpression)rightExpression
{
    return _right;
}

- (CPPredicateOperatorType)predicateOperatorType
{
    return _type;
}

- (CPComparisonPredicateModifier)comparisonPredicateModifier
{
    return _modifier;
}

- (unsigned)options
{
    return _options;
}

- (SEL)customSelector
{
    return _customSelector;
}

- (BOOL)_evaluateValue:leftResult rightValue:rightResult
{
    var CPDiacriticInsensitiveSearch = 128; // this has to be declared in CPString rather than here
    
    var leftIsNil = (leftResult == nil || [leftResult isEqual:[CPNull null]]);
    var rightIsNil = (rightResult == nil || [rightResult isEqual:[CPNull null]]);

    if (leftIsNil || rightIsNil)
        return ((leftIsNil == rightIsNil) &&
            (_type == CPEqualToPredicateOperatorType ||
            _type == CPLessThanOrEqualToPredicateOperatorType ||
            _type == CPGreaterThanOrEqualToPredicateOperatorType));
        
    var string_compare_options = 0;

    // left and right should be casted first [CAST()] following 10.5 rules.
    switch (_type)
    {
        case CPLessThanPredicateOperatorType:
            return ([leftResult compare:rightResult] == CPOrderedAscending);

        case CPLessThanOrEqualToPredicateOperatorType:
            return ([leftResult compare:rightResult] != CPOrderedDescending);

        case CPGreaterThanPredicateOperatorType:
            return ([leftResult compare:rightResult] == CPOrderedDescending);

        case CPGreaterThanOrEqualToPredicateOperatorType:
            return ([leftResult compare:rightResult] != CPOrderedAscending);

        case CPEqualToPredicateOperatorType:
            return [leftResult isEqual:rightResult];

        case CPNotEqualToPredicateOperatorType:
            return (![leftResult isEqual:rightResult]);

        case CPMatchesPredicateOperatorType:
           if (_options & CPDiacriticInsensitivePredicateOption)
                {    
                  leftResult = _stripDiacritics(leftResult);
                  rightResult = _stripDiacritics(rightResult);
                }
    
            var commut = (_options & CPCaseInsensitivePredicateOption) ? "gi":"g";
            var reg = new RegExp(rightResult,commut);
            return reg.test(leftResult);

        case CPLikePredicateOperatorType: 
            if (_options & CPDiacriticInsensitivePredicateOption)
                {    
                  leftResult = _stripDiacritics(leftResult);
                  rightResult = _stripDiacritics(rightResult);
                }
            var commut = (_options & CPCaseInsensitivePredicateOption) ? "gi":"g";
            var like_pattern = escape_regex(rightResult);
            var reg = new RegExp(like_pattern,commut);
            return reg.test(leftResult);

        case CPBeginsWithPredicateOperatorType:
            var range = CPMakeRange(0,[rightResult length]);
            if (_options & CPCaseInsensitivePredicateOption) string_compare_options |= CPCaseInsensitiveSearch;
            if (_options & CPDiacriticInsensitivePredicateOption) string_compare_options |= CPDiacriticInsensitiveSearch;

            return ([leftResult compare:rightResult options:string_compare_options range:range] == CPOrderedSame);

        case CPEndsWithPredicateOperatorType:
            var range = CPMakeRange([leftResult length] - [rightResult length],[rightResult length]);
            if (_options & CPCaseInsensitivePredicateOption) string_compare_options |= CPCaseInsensitiveSearch;
            if (_options & CPDiacriticInsensitivePredicateOption) string_compare_options |= CPDiacriticInsensitiveSearch;

            return ([leftResult compare:rightResult options:string_compare_options range:range] == CPOrderedSame);

        case CPInPredicateOperatorType:
         // Handle special case where rightResult is a collection and leftResult an element of it.
         if (![rightResult isKindOfClass: [CPString class]])
           {
             if (![rightResult respondsToSelector: @selector(objectEnumerator)])
                 [CPException raise:CPInvalidArgumentException reason:@"The right hand side for an IN operator must be a collection"];
             
             var e = [rightResult objectEnumerator],
                 value;
             while (value = [e nextObject])
                 if ([value isEqual:leftResult]) 
                   return YES;

             return NO;
           }

        if (_options & CPCaseInsensitivePredicateOption) string_compare_options |= CPCaseInsensitiveSearch;
        if (_options & CPDiacriticInsensitivePredicateOption) string_compare_options |= CPDiacriticInsensitiveSearch;

         return ([rightResult rangeOfString:leftResult options:string_compare_options].location != CPNotFound);

        case CPCustomSelectorPredicateOperatorType:
            return [leftResult performSelector:_customSelector withObject:rightResult];

        default:
            return NO;
    }
}

- (BOOL)evaluateWithObject:(id)object substitutionVariables:(CPDictionary)variables
{
    var p = [self predicateWithSubstitutionVariables:variables];
    return [p evaluateWithObject:object];
}

- (BOOL)evaluateWithObject:(id)object
{
    var leftValue = [_left expressionValueWithObject:object context:nil];
    var rightValue = [_right expressionValueWithObject:object context:nil];

    if (_modifier == CPDirectPredicateModifier)
        return [self _evaluateValue:leftValue rightValue:rightValue];
    else
    {
        if (![leftValue respondsToSelector:@selector(objectEnumerator)])
            [CPException raise:CPInvalidArgumentException reason:@"The left hand side for an ALL or ANY operator must be either a CPArray or a CPSet"];

    var e = [leftValue objectEnumerator],
        result = (_modifier == CPAllPredicateModifier),
        value;

    while (value = [e nextObject])
        {
            var eval = [self _evaluateValue:value rightValue:rightValue];
            if (eval != result) 
                return eval;
        }

    return result;
    }
}

@end

var _stripDiacritics = function _stripDiacritics(str)
{
    var output = "";

    var diacritics = [[192,198],[224,230],[231,231],[232,235],[236,239],[242,246],[249,252]];    
    var normalized = [65,97,99,101,105,111,117];
 
     for (var indexSource = 0; indexSource < [str length]; indexSource++)
    {
        var code = str.charCodeAt(indexSource);

        for (var i = 0; i < diacritics.length; i++)
        {
            var drange = diacritics[i];         
            
            if (code >= drange[0] && code <= drange[1])
            {
                code = normalized[i];
                break;
            }
        }
        
        output += String.fromCharCode(code);
    }
            
    return output;
}

var escape_regex = function escape_regex(str)
{
    var regexchar = ['(',')','{','}','.','*','+','?','|','^','$'];
    var replace = ['\(','\)','\{','\}','\.','.*','\+','.?','\|','\^','\$'];
    
    for (var i = 0; i < regexchar.length; i++)
        str = str.replace(regexchar[i], replace[i]);
        
    return str;    
}
