
@import "CPString.j"
@import "CPArray.j"
@import "CPDictionary.j"

@import "CPExpression.j"

@implementation CPExpression_function : CPExpression
{
   CPString _name;
   SEL      _selector;
   CPArray  _arguments;
   int      _argc;
   CPInvocation _invocation;
}

- (void)setSelector:(SEL)selector
{
  _selector = selector;
 /*
  var signature = [CPExpression_function methodSignatureForSelector:selector];
  _invocation = [CPInvocation invocationWithMethodSignature:signature];
  [_invocation setSelector:_selector];
  [_invocation setTarget:self];
 */
}

- (id)initWithName:(CPString)name arguments:(CPArray)arguments
{
    [super initWithExpressionType:CPFunctionExpressionType];  
     
    _name = name;
    _arguments = arguments;
    _argc = [_arguments count];
    return self;
}

- (id)initWithCoder:(CPCoder)coder
{
    var name = [coder decodeObjectForKey:@"CPExpressionFunctionName"],
        arguments = [coder decodeObjectForKey:@"CPExpressionFunctionArguments"];
    
    return [self initWithName:name arguments:arguments];
}

- (void)encodeWithCoder:(CPCoder)coder
{
    [coder encodeObject:_name forKey:@"CPExpressionFunctionName"];
    [coder encodeObject:_arguments forKey:@"CPExpressionArguments"];
}

- (CPString)function
{
   return _name;
}

- (CPArray)arguments
{
   return _arguments;
}

- (id)expressionValueWithObject:(id)object context:(CPDictionary)context
{
    var eval_args = [CPArray array],
        i;

  for (i = 0; i < _argc; i++)
    {
      var arg = [[_arguments objectAtIndex:i] expressionValueWithObject:object context:context];
      if(arg != nil) [eval_args addObject:arg];
    }

  return [self performSelector:_selector withObject:eval_args];
}

- (CPString)description
{  
   var  result =  [CPString stringWithFormat:@"%s(", _name],
        i;
  
   for (i = 0; i < _argc; i++){
    result = result + [CPString stringWithFormat:@"%s%s", [_arguments objectAtIndex:i] , (i+1<_argc) ? ", " : ""];
    }
    
   result = result + ")";
   
   return result ;
}

- (CPExpression)_expressionWithSubstitutionVariables:(CPDictionary)variables
{
   var  array = [CPArray array],
        i;
    
   for (i = 0; i < _argc; i++)
    [array addObject:[[_arguments objectAtIndex:i] _expressionWithSubstitutionVariables:variables]];

   return [CPExpression expressionForFunction:_name arguments:array];   
}

/*
10.5 documentation for expressionForFunction:arguments:

 This method throws an exception immediately if the selector is unknown; it throws at runtime if the parameters are incorrect.
 Parameters
 name
 The name of the function to invoke.
 
 parameters
 An array containing NSExpression objects that will be used as parameters during the invocation of selector.
 
 For a selector taking no parameters, the array should be empty. For a selector taking one or more parameters, the array should contain one NSExpression object which will evaluate to an instance of the appropriate type for each parameter.
 
 If there is a mismatch between the number of parameters expected and the number you provide during evaluation, an exception may be raised or missing parameters may simply be replaced by nil (which occurs depends on how many parameters are provided, and whether you have over- or underflow).
 
 Return Value
 A new expression that invokes the function name using the parameters in parameters.
*/

// 10.4 Functions are sum, count, min, max, average


- (CPNumber)average:(CPArray)parameters 
{
  if(_argc < 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var i,
      sum = 0.0;
    
  for (i = 0; i < _argc; i++)
    {
      sum += [[parameters objectAtIndex: i] doubleValue];
    }
  return [CPNumber numberWithDouble: sum / _argc];
}

- (CPNumber)sum:(CPArray)parameters
{
  if(_argc < 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var i,
      sum = 0.0;
    
  for (i = 0; i < _argc; i++)
    {
      sum += [[parameters objectAtIndex: i] doubleValue];
    }
  return [CPNumber numberWithDouble: sum];
}

- (CPNumber)count:(CPArray)parameters
{
  if(_argc < 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  return [CPNumber numberWithUnsignedInt: [[parameters objectAtIndex: 0] count]];
}

- (CPNumber)min:(CPArray)parameters
{
  if(_argc < 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   i,
        min = 0.0,
        cur;
  
  if (_argc > 0)
    {
      min = [[parameters objectAtIndex: 0] doubleValue];
    }

  for (i = 1; i < _argc; i++)
    {
      cur = [[parameters objectAtIndex: i] doubleValue];
      if (min > cur)
        {
          min = cur;
        }
    }
  return [CPNumber numberWithDouble: min];
}

- (CPNumber)max:(CPArray)parameters
{
  if(_argc < 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   i,
        max = 0.0,
        cur;
  
  if (_argc > 0)
      max = [[parameters objectAtIndex: 0] doubleValue];
    
  for (i = 1; i < _argc; i++)
    {
      cur = [[parameters objectAtIndex: i] doubleValue];
      if (max < cur)
           max = cur;
     }
  return [CPNumber numberWithDouble: max];
}

- (CPNumber)add:to:(CPArray)parameters
{
  if(_argc != 2)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   left = [parameters objectAtIndex: 0],
        right = [parameters objectAtIndex: 1];

  return [CPNumber numberWithDouble: [left doubleValue] + [right doubleValue]];
}

- (CPNumber)from:subtract:(CPArray)parameters
{
  if(_argc != 2)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   left = [parameters objectAtIndex: 0],
        right = [parameters objectAtIndex: 1];

  return [CPNumber numberWithDouble: [left doubleValue] - [right doubleValue]];
}

- (CPNumber)multiply:by:(CPArray)parameters
{
  if(_argc != 2)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   left = [parameters objectAtIndex: 0],
        right = [parameters objectAtIndex: 1];
  
  return [CPNumber numberWithDouble: [left doubleValue] * [right doubleValue]];
}

- (CPNumber)divide:by:(CPArray)parameters
{
  if(_argc != 2)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var   left = [parameters objectAtIndex: 0],
        right = [parameters objectAtIndex: 1];

  return [CPNumber numberWithDouble: [left doubleValue] / [right doubleValue]];
}

- (CPNumber)sqrt:(CPArray)parameters
{
  if(_argc != 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

  var num = [[parameters objectAtIndex: 0] doubleValue];

  return [CPNumber numberWithDouble: Math.sqrt(num)];
}

- (CPNumber)abs:(CPArray)parameters
{
  if(_argc != 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

    var num = [[parameters objectAtIndex: 0] doubleValue];
    
    return [CPNumber numberWithDouble:Math.abs(num)];
}

- (CPDate)now
{
    return [CPDate date];
}

- (CPNumber)ln:(CPArray)parameters
{
  if(_argc != 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

    var num = [[parameters objectAtIndex: 0] doubleValue];
    
    return [CPNumber numberWithDouble:Math.log(num)];
}

- (CPNumber)exp:(CPArray)parameters
{
  if(_argc != 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

    var num = [[parameters objectAtIndex: 0] doubleValue];
    
    return [CPNumber numberWithDouble:Math.exp(num)];
}

- (CPNumber)ceiling:(CPArray)parameters
{
  if(_argc != 1)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

    var num = [[parameters objectAtIndex: 0] doubleValue];
    
    return [CPNumber numberWithDouble:Math.ceil(num)];
}

- (CPNumber)random
{
    return [CPNumber numberWithDouble:Math.random()];
}

- (CPNumber)modulus:by:(CPArray)parameters
{
  if(_argc != 2)
    [CPException raise:CPInvalidArgumentException reason:"Invalid number of parameters"];

    var left = [parameters objectAtIndex: 0],
        right = [parameters objectAtIndex: 1];

    return [CPNumber numberWithInt:([left intValue] % [right intValue])];
}


/*
- (CPNumber)median:(CPArray)parameters
{
}
- (CPNumber)mode:(CPArray)parameters
{
}
- (CPNumber)stddev:(CPArray)parameters
{
}
- (CPNumber)log:(CPArray)parameters
{
}
- (CPNumber)raise:to:(CPArray)parameters
{
}

- (CPNumber)trunc:(CPArray)parameters
{
}
- (CPNumber)random:(CPArray)parameters
{
}

// These functions are generated when parsing

- (id) first: (CPArray)parameters
{
  return [[parameters objectAtIndex: 0] objectAtIndex: 0];
}

- (id) last: (CPArray)parameters
{
  return [[parameters objectAtIndex: 0] lastObject];
}

- (CPNumber)chs:(CPArray)parameters
{
  return [CPNumber numberWithInt: -[[parameters objectAtIndex: 0] intValue]];
}

- (id)index:(CPArray)parameters
{
  var left = [parameters objectAtIndex: 0];
  var right = [parameters objectAtIndex: 1];

  if ([left isKindOfClass: [CPDictionary class]])
    {
      return [left objectForKey: right];
    }
  else
    {
      // raises exception if invalid
      return [left objectAtIndex: [right unsignedIntValue]];
    }
}
*/

@end
